from os.path import join

from kash.config.logger import get_logger
from kash.errors import InvalidInput
from kash.exec import kash_action
from kash.exec.preconditions import has_html_body, has_text_body
from kash.kits.media.libs.pdf_output import html_to_pdf
from kash.llm_utils.clean_headings import clean_heading
from kash.model import FileExt, Format, Item, ItemType
from kash.utils.common.format_utils import fmt_loc
from kash.workspaces import current_ws

log = get_logger(__name__)


@kash_action(
    precondition=has_text_body | has_html_body,
    mcp_tool=True,
)
def create_pdf(item: Item) -> Item:
    """
    Create a PDF from text or Markdown.
    """
    if not item.body:
        raise InvalidInput(f"Item must have a body: {item}")

    pdf_item = item.derived_copy(type=ItemType.export, format=Format.pdf, file_ext=FileExt.pdf)
    pdf_path, _found, _old_pdf_path = current_ws().store_path_for(pdf_item)
    log.message("Will save PDF to: %s", fmt_loc(pdf_path))
    base_dir = current_ws().base_dir
    full_pdf_path = join(base_dir, pdf_path)

    clean_title = clean_heading(item.abbrev_title())

    # Convert to HTML if necessary.
    if item.format == Format.html:
        content_html = f"""
            <h1>{clean_title}</h1>
            {item.body_text()}
            """
    else:
        content_html = f"""
            <h1>{clean_title}</h1>
            {item.body_as_html()}
            """

    # Add directly to the store.
    html_to_pdf(
        content_html,
        full_pdf_path,
        title=item.title,
    )
    pdf_item.external_path = full_pdf_path

    return pdf_item



from kash.config.logger import get_logger
from kash.errors import InvalidInput
from kash.exec import kash_action
from kash.exec.preconditions import is_url_item
from kash.media_base.media_services import canonicalize_media_url, list_channel_items
from kash.model import NO_ARGS, ActionInput, ActionResult, Item

log = get_logger(__name__)


@kash_action(
    expected_args=NO_ARGS,
    precondition=is_url_item,
)
def list_channel(input: ActionInput) -> ActionResult:
    """
    List the contents of a media channel (YouTube, Apple Podcasts, etc.) channel, saving
    the URL of every audio or video as a resource item. Only adds the resources.
    Does not download any media.
    """
    item = input.items[0]
    if not item.url:
        raise InvalidInput("Item must have a URL")
    if not canonicalize_media_url(item.url):
        raise InvalidInput("Media format not supported")

    metadata_list = list_channel_items(item.url)

    result_items = []
    for metadata in metadata_list:
        if not canonicalize_media_url(metadata.url):
            log.warning("Skipping non-recognized video URL: %s", metadata.url)
            continue

        item = Item.from_media_metadata(metadata)
        result_items.append(item)

    return ActionResult(result_items)



from chopdiff.transforms import WINDOW_128_PARA, adds_headings

from kash.config.logger import get_logger
from kash.exec import kash_action, llm_transform_item
from kash.model import LLM, Item, LLMOptions, Message, MessageTemplate

log = get_logger(__name__)


llm_options = LLMOptions(
    model=LLM.default_fast,
    diff_filter=adds_headings,
    windowing=WINDOW_128_PARA,
    system_message=Message(
        """
        You are a careful and precise editor.
        You give exactly the results requested without additional commentary.
        """
    ),
    body_template=MessageTemplate(
        """
        Insert headings into the following text using `<h2>` tags. 

        - Add a heading every time topics change, typically after 3-6 paragraphs, but follow your
          best judgement in terms of when the change in topic occurs.

        - Each heading should describe what is covered by the paragraphs that follow.

        - DO NOT change any text other than to add headings, each on its own line, in
          between the paragraphs of the text.
                        
        - Section headings should be concise and specific. For example, use
          "Importance of Sleep" and not just "Sleep", or "Reflections on Johanna's Early Childhood" and
          not just "Childhood".
          
        - Do NOT give any introductory response at the beginning, such as "Here is the text
          with headings added".

        - If the input is short, you can add a single heading at the beginning.

        - If the input is very short or unclear, output the text exactly, without adding any headings.

        Input text:

        {body}

        Output text (identical to input, but with headings added):
        """
    ),
)


@kash_action(llm_options=llm_options)
def insert_section_headings(item: Item) -> Item:
    """
    Insert headings into a text as <h2> tags.
    """
    return llm_transform_item(item)



from kash.exec.precondition_registry import kash_precondition
from kash.model.items_model import Item


@kash_precondition
def has_video_id(item: Item) -> bool:
    from kash.media_base.media_services import get_media_id

    return bool(item.url and get_media_id(item.url))


@kash_precondition
def is_youtube_video(item: Item) -> bool:
    from kash.kits.media.libs import youtube

    return bool(item.url and youtube.canonicalize(item.url))



from pathlib import Path

from kash.exec import import_action_subdirs

# This hook can be used for auto-registering actions from any module.
import_action_subdirs(["actions"], __package__, Path(__file__).parent)



import logging
import os
import platform
from datetime import datetime
from logging import ERROR
from textwrap import dedent

from strif import atomic_output_file

from kash.config.logger import get_logger
from kash.config.settings import APP_NAME
from kash.kits.media.libs import media_templates_dir
from kash.web_gen import base_templates_dir as common_templates_dir
from kash.web_gen.template_render import render_web_template

log = get_logger(__name__)


def weasyprint_setup() -> None:
    is_macos = platform.system() == "Darwin"
    if is_macos:
        dylib = os.getenv("DYLD_LIBRARY_PATH")
        homebrew_lib = "/opt/homebrew/lib"

        new_dylib = f"{dylib}:{homebrew_lib}" if dylib else homebrew_lib
        print("Setting DYLD_LIBRARY_PATH to: %s", new_dylib)
        # os.environ["DYLD_LIBRARY_PATH"] = new_dylib

    try:
        import weasyprint  # noqa: F401
    except ImportError:
        if is_macos:
            log.exception(
                dedent(
                    """Failed to import weasyprint. Installing with `brew` may help. See:
                    https://github.com/Kozea/WeasyPrint/issues/1448
                    https://github.com/astral-sh/uv/issues/6971
                    """
                )
            )

    # Reduce weasyprint noisiness.
    log_levels = {
        "weasyprint": ERROR,
        "weasyprint.progress": ERROR,
    }
    for logger_name, level in log_levels.items():
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)
        logger.propagate = True


def html_to_pdf(
    html_content: str,
    output_file_path: str,
    title: str | None = None,
) -> None:
    """
    Converts an HTML doc to a nicely formatted PDF file.
    """

    today = datetime.now().strftime("%Y-%m-%d")
    footer = f"Generated by {APP_NAME} on {today}"

    title = title or ""

    # Scale the content down a bit.
    scaled_html = f"""
        <style>
            body {{
                transform: scale(0.9);
                transform-origin: top left;
            }}
        </style>
        {html_content}
        """

    extra_css = render_web_template(
        media_templates_dir,
        "pdf_styles.css.jinja",
        {
            "title": title,
            "footer": footer,
        },
    )

    full_html = render_web_template(
        common_templates_dir,
        "base_webpage.html.jinja",
        {
            "title": title,
            "content": scaled_html,  # Use the scaled content
            "extra_css": extra_css,
        },
    )

    # Create PDF.
    weasyprint_setup()
    import weasyprint

    weasy_html = weasyprint.HTML(string=full_html)
    with atomic_output_file(output_file_path, make_parents=True) as temp_file_path:
        weasy_html.write_pdf(temp_file_path)



import json

from kash.config.logger import get_logger
from kash.errors import ApiResultError, InvalidInput
from kash.exec import kash_action
from kash.exec.preconditions import has_html_body, has_text_body
from kash.kits.media.libs.speaker_labels import find_speaker_labels
from kash.llm_utils.fuzzy_parsing import fuzzy_parse_json
from kash.llm_utils.llm_completion import llm_template_completion
from kash.media_base.timestamp_citations import html_speaker_id_span
from kash.model import LLM, Item, ItemType, Message, MessageTemplate
from kash.utils.common.string_replace import replace_multiple

log = get_logger(__name__)


@kash_action(
    precondition=has_text_body | has_html_body,
)
def identify_speakers(item: Item) -> Item:
    """
    Identify speakers in a transcript and replace placeholders with their names.
    """
    if not item.body:
        raise InvalidInput("Item must have a body")

    # Find all speaker labels and their offsets
    speaker_labels = find_speaker_labels(item.body)
    if not speaker_labels:
        log.warning("This document has no speaker labels! Skipping this action.")
        return item  # No changes needed.

    # Prepare the system message and template for LLM.
    system_message = Message("You are an assistant that identifies speakers in transcripts.")
    identification_template = MessageTemplate(
        """
        The transcript below includes speakers identified by IDs like 'SPEAKER 0' or 'SPEAKER 1'.
        Based on the transcript, provide a mapping from speaker IDs to actual speaker names.
        The mapping should be in JSON format like {{"0": "Alice", "1": "Bob"}}.
        If you are not sure from the content, leave the names as is, writing something like
        {{"0": "Alice", "1": "SPEAKER 1"}} or {{"0": "SPEAKER 0", "1": "SPEAKER 1"}}.
        Transcript:
        {body}
        """
    )

    # Perform LLM completion to get the speaker mapping.
    mapping_str = llm_template_completion(
        model=LLM.gpt_4o_mini,
        system_message=system_message,
        input=item.body,
        body_template=identification_template,
    ).content

    # Parse the mapping.
    try:
        speaker_mapping = fuzzy_parse_json(mapping_str)
        if not speaker_mapping:
            log.error("Could not parse speaker mapping: %s", mapping_str)
            raise ApiResultError("Could not parse speaker mapping")
        log.message("Identified speakers from transcript: %s", speaker_mapping)
    except json.JSONDecodeError as e:
        raise ApiResultError(f"Failed to parse speaker mapping from LLM output: {e}")

    # Prepare replacements.
    replacements = []
    for match in speaker_labels:
        speaker_id = match.attribute_value
        if not speaker_id:
            raise InvalidInput(f"Speaker id not found: {match}")
        new_speaker_name = speaker_mapping.get(speaker_id, f"SPEAKER {speaker_id}")
        # Prepare replacement text.
        new_span = html_speaker_id_span(f"**{new_speaker_name}:**", speaker_id)
        replacements.append((match.start_offset, match.end_offset, new_span))

    # Perform replacements.
    updated_body = replace_multiple(item.body, replacements)

    result_item = item.derived_copy(type=ItemType.doc, body=updated_body)
    return result_item



from kash.config.logger import get_logger
from kash.exec import kash_action
from kash.exec.preconditions import is_audio_resource, is_url_item, is_video_resource
from kash.kits.media.actions.add_description import add_description
from kash.kits.media.actions.add_summary_bullets import add_summary_bullets
from kash.kits.media.actions.caption_paras import caption_paras
from kash.kits.media.actions.insert_frame_captures import insert_frame_captures
from kash.kits.media.actions.insert_section_headings import insert_section_headings
from kash.kits.media.actions.transcribe_format import transcribe_format
from kash.model import Item

log = get_logger(__name__)


@kash_action(
    precondition=is_url_item | is_audio_resource | is_video_resource,
    mcp_tool=True,
)
def transcribe_annotate_summarize(item: Item) -> Item:
    """
    A fancy action to transcribe a video, format the transcript into paragraphs,
    backfill timestamps, and add a summary and description.
    """
    formatted = transcribe_format(item)

    with_headings = insert_section_headings(formatted)

    with_captions = caption_paras(with_headings)

    with_summary = add_summary_bullets(with_captions)

    with_description = add_description(with_summary)

    with_frames = insert_frame_captures(with_description)

    return with_frames



from pathlib import Path
from typing import Tuple

import click

from kash.actions.core.strip_html import strip_html
from kash.actions.core.webpage_config import webpage_config
from kash.actions.core.webpage_generate import webpage_generate
from kash.exec import assemble_action_input
from kash.kits.media.actions.add_description import add_description
from kash.kits.media.actions.add_summary_bullets import add_summary_bullets
from kash.kits.media.actions.backfill_timestamps import backfill_timestamps
from kash.kits.media.actions.break_into_paragraphs import break_into_paragraphs
from kash.kits.media.actions.caption_paras import caption_paras
from kash.kits.media.actions.insert_frame_captures import insert_frame_captures
from kash.kits.media.actions.insert_section_headings import insert_section_headings
from kash.kits.media.actions.transcribe import transcribe
from kash.model import ActionInput, Item
from kash.workspaces import get_workspace


def basic_transcribe(item: Item) -> Item:
    # These are all simple actions, so take just accept and return one Item.
    transcribed = transcribe(item)

    stripped = strip_html(transcribed)

    paragraphs = break_into_paragraphs(stripped)

    clean_transcript = backfill_timestamps(paragraphs)

    return clean_transcript


def annotate_transcription(item: Item) -> Item:
    with_headings = insert_section_headings(item)

    with_captions = caption_paras(with_headings)

    with_summary = add_summary_bullets(with_captions)

    with_description = add_description(with_summary)

    md_final = insert_frame_captures(with_description)

    return md_final


def deep_transcribe(url: str, workspace_path: Path, rerun: bool) -> Tuple[Path, Path]:

    # Get or initialize a workspace directory.
    ws = get_workspace(workspace_path)

    input = assemble_action_input(ws, url)

    clean_transcript = basic_transcribe(input.items[0])

    annotated_transcript = annotate_transcription(clean_transcript)

    # These are regular actions that require ActionInput/ActionResult.
    config = webpage_config(ActionInput(items=[annotated_transcript]))

    html_final = webpage_generate(ActionInput(items=config.items))

    assert annotated_transcript.store_path
    assert html_final.items[0].store_path

    md_path = Path(annotated_transcript.store_path)
    html_path = Path(html_final.items[0].store_path)

    print("\n\nAll done!")
    print(f"Cleanly formatted Markdown/HTML is at: {md_path}")
    print(f"Browser-ready HTML is at: {html_path}")

    return md_path, html_path


@click.command()
@click.argument("url", type=click.STRING)
@click.option(
    "--workspace",
    type=click.STRING,
    help="The workspace directory to use for files, metadata, and cache",
    default=".",
)
@click.option(
    "--rerun",
    is_flag=True,
    help="Rerun actions even if the outputs already exist",
)
def cli_main(url: str, workspace: str, rerun: bool) -> None:
    """
    Take a video or audio URL (such as YouTube), download and cache it, and
    perform a "deep transcription" of it, including, full transcription,
    identifying speakers, adding sections, timestamps, and annotations,
    and inserting frame captures.
    """
    deep_transcribe(url, Path(workspace), rerun)


if __name__ == "__main__":
    cli_main()



from chopdiff.html.html_in_md import div_wrapper

from kash.actions.core.summarize_as_bullets import summarize_as_bullets
from kash.config.logger import get_logger
from kash.exec import kash_action
from kash.exec.preconditions import is_text_doc
from kash.model import Format, Item, ItemType
from kash.utils.common.type_utils import not_none

log = get_logger(__name__)


SUMMARY = "summary"
"""Class name for the summary."""

ORIGINAL = "original"
"""Class name for the original content."""


@kash_action(
    precondition=is_text_doc,
)
def add_summary_bullets(item: Item) -> Item:
    """
    Add a summary of the content (from `summarize_as_bullets`) above the full text of the item,
    with each wrapped in a div.
    """
    summary_item = summarize_as_bullets(item)

    wrap_summary = div_wrapper(class_name=SUMMARY)
    wrap_original = div_wrapper(class_name=ORIGINAL)

    combined_body = (
        wrap_summary(not_none(summary_item.body)) + "\n\n" + wrap_original(not_none(item.body))
    )

    output_item = item.derived_copy(
        type=ItemType.doc,
        format=Format.md_html,
        body=combined_body,
    )

    return output_item



import os
import tempfile
from datetime import date
from pathlib import Path
from typing import Any

import yt_dlp
from frontmatter_format import to_yaml_string

from kash.config.logger import get_logger
from kash.errors import ApiResultError
from kash.utils.common.url import Url
from kash.utils.file_utils.file_formats_model import MediaType

log = get_logger(__name__)


def parse_date(upload_date: str | date) -> date:
    if isinstance(upload_date, str):
        return date.fromisoformat(upload_date)
    elif isinstance(upload_date, date):
        return upload_date
    raise ValueError(f"Invalid date: {upload_date}")


def ydl_extract_info(url: Url) -> dict[str, Any]:
    ydl_opts = {
        "extract_flat": "in_playlist",  # Extract metadata only, without downloading.
        "quiet": True,
        "dump_single_json": True,
        "logger": log,
    }
    with yt_dlp.YoutubeDL(ydl_opts) as ydl:
        result = ydl.extract_info(str(url), download=False)

        log.save_object("yt_dlp result", None, to_yaml_string(result, stringify_unknown=True))

        if not isinstance(result, dict):
            raise ApiResultError(f"Unexpected result from yt_dlp: {result}")

        return result


def ydl_download_media(
    url: Url, target_dir: Path | None = None, media_types: list[MediaType] | None = None
) -> dict[MediaType, Path]:
    """
    Download and convert to mp3 and mp4 using yt_dlp, which is generally the best
    library for this.
    """

    if not media_types:
        media_types = [MediaType.audio, MediaType.video]

    temp_dir = target_dir or tempfile.mkdtemp()
    ydl_opts: dict[str, Any]
    if MediaType.video in media_types:
        ydl_opts = {
            # Try for best video+audio, fall back to best available.
            # Might want to support smaller sizes tho.
            # "format": "bestvideo[height<=720]+bestaudio/best",
            "format": "bestvideo+bestaudio/best",
            "outtmpl": os.path.join(temp_dir, "media.%(id)s.%(ext)s"),
            "postprocessors": [
                {
                    "key": "FFmpegVideoConvertor",
                    "preferedformat": "mp4",  # Yep, it's really spelled this way in yt_dlp.
                },
                {
                    "key": "FFmpegExtractAudio",
                    "preferredcodec": "mp3",
                    "preferredquality": "192",
                },
            ],
            "keepvideo": True,  # By default yt_dlp will delete the video file.
        }
    else:
        ydl_opts = {
            # Try for best video+audio, fall back to best available.
            "format": "bestaudio/best",
            "outtmpl": os.path.join(temp_dir, "media.%(id)s.%(ext)s"),
            "postprocessors": [
                {
                    "key": "FFmpegExtractAudio",
                    "preferredcodec": "mp3",
                    "preferredquality": "192",
                },
            ],
        }

    # Use our logger.
    ydl_opts["logger"] = log  # pylance: ignore

    log.info("Extracting media from %s at %s using ydl_opts: %s", url, temp_dir, ydl_opts)

    info_dict = None
    with yt_dlp.YoutubeDL(ydl_opts) as ydl:
        info_dict = ydl.extract_info(url, download=True)
        media_file_path = ydl.prepare_filename(info_dict)

    result_paths = {}

    log.info("ydl output filename: %s", media_file_path)

    # Check audio and video files exist.
    if MediaType.audio in media_types:
        mp3_path = os.path.splitext(media_file_path)[0] + ".mp3"
        if os.path.exists(mp3_path):
            result_paths[MediaType.audio] = Path(mp3_path)
        else:
            log.warning("mp3 download not found: %s", mp3_path)
    if MediaType.video in media_types:
        mp4_path = os.path.splitext(media_file_path)[0] + ".mp4"
        if os.path.exists(mp4_path):
            result_paths[MediaType.video] = Path(mp4_path)
        else:
            log.warning("mp4 download not found: %s", mp4_path)

    return result_paths



from chopdiff.docs import TextDoc
from chopdiff.transforms import WINDOW_2K_WORDTOKS, changes_whitespace

from kash.config.logger import get_logger
from kash.exec import SkipItem, kash_action, llm_transform_item
from kash.model import LLM, Item, LLMOptions, Message, MessageTemplate, Param

log = get_logger(__name__)


llm_options = LLMOptions(
    model=LLM.default_fast,
    diff_filter=changes_whitespace,
    windowing=WINDOW_2K_WORDTOKS,
    system_message=Message(
        """
        You are a careful and precise editor.
        You give exactly the results requested without additional commentary.
        """
    ),
    body_template=MessageTemplate(
        """
        Format this text according to these rules:

        - Break the following text into paragraphs so it is readable and organized.

        - Add a paragraph break whenever the topic changes.

        - Paragraphs can be short or up to several sentences long.

        - Do *not* change any words of the text. Add line breaks only.

        - Preserve all Markdown formatting.

        - ONLY GIVE THE FORMATTED TEXT, with no other commentary.

        Original text:

        {body}

        Formatted text:
        """
    ),
)


@kash_action(
    llm_options=llm_options,
    params=(
        Param(
            "max_sent_per_para",
            "Maximum number of sentences per paragraph.",
            type=int,
            default_value=7,
        ),
    ),
)
def break_into_paragraphs(item: Item, max_sent_per_para: int = 7) -> Item:
    """
    Reformat text as paragraphs. Preserves all text exactly except for whitespace changes.
    Does nothing if there are very few sentences.
    """
    if not item.body:
        raise SkipItem()

    # Check each paragraph's sentence count to see if we can skip this step altogether.
    doc = TextDoc.from_text(item.body)
    log.message("Doc size: %s", doc.size_summary())
    biggest_para = max(doc.paragraphs, key=lambda p: len(p.sentences))
    can_skip = len(biggest_para.sentences) <= max_sent_per_para

    log.message(
        "Checking if we need to break into paragraphs: "
        "biggest paragraph has %d sentences vs max of %d so %s",
        len(biggest_para.sentences),
        max_sent_per_para,
        "can skip" if can_skip else "must run",
    )
    if can_skip:
        raise SkipItem()

    # Finally, do the transform.
    result_item = llm_transform_item(item)
    return result_item



from dataclasses import asdict, dataclass

from frontmatter_format import to_yaml_string

from kash.config.logger import get_logger
from kash.errors import InvalidInput
from kash.kits.media.libs import media_templates_dir
from kash.kits.media.libs.media_preconditions import is_youtube_video
from kash.llm_utils.clean_headings import clean_heading, summary_heading
from kash.media_base.media_services import get_media_id
from kash.model.items_model import Item, ItemType
from kash.utils.common.type_utils import as_dataclass
from kash.utils.file_utils.file_formats_model import Format
from kash.web_gen import base_templates_dir
from kash.web_gen.template_render import render_web_template
from kash.workspaces.source_items import find_upstream_item

log = get_logger(__name__)


@dataclass
class VideoInfo:
    youtube_id: str
    title: str
    description: str
    topics: list[str]


@dataclass
class VideoGallery:
    title: str
    videos: list[VideoInfo]


def video_gallery_config(items: list[Item]) -> Item:
    """
    Get an item with the config for a video gallery.
    """
    videos = []
    for item in items:
        source_item = find_upstream_item(item, is_youtube_video)

        youtube_id = get_media_id(source_item.url)
        log.message(
            "Pulling video from source item: %s: %s", source_item.store_path, source_item.url
        )
        if not youtube_id or not is_youtube_video(source_item):
            raise InvalidInput(f"Item must be a YouTube URL with id: {source_item}")

        video_info = VideoInfo(
            youtube_id=youtube_id,
            title=clean_heading(item.abbrev_title()),
            description=item.abbrev_description(),
            topics=[],  # TODO
        )
        videos.append(video_info)

    title = summary_heading([video.title for video in videos])
    gallery = VideoGallery(title=title, videos=videos)

    config_item = Item(
        title=f"Config for {title}",
        type=ItemType.config,
        format=Format.yaml,
        body=to_yaml_string(asdict(gallery)),
    )
    return config_item


def video_gallery_generate(config_item: Item) -> str:
    """
    Generate a video gallery web page using the supplied config.
    """
    config = config_item.read_as_config()
    video_gallery = as_dataclass(config, VideoGallery)  # Checks the format.

    content = render_web_template(
        media_templates_dir,
        "youtube_gallery.html.jinja",
        asdict(video_gallery),
    )

    return render_web_template(
        base_templates_dir,
        "base_webpage.html.jinja",
        {"title": video_gallery.title, "content": content},
    )



from datetime import date
from pathlib import Path
from typing import Any
from urllib.parse import parse_qs, urlparse

from frontmatter_format import to_yaml_string
from typing_extensions import override
from yt_dlp.utils import DownloadError

from kash.config.logger import get_logger
from kash.config.text_styles import EMOJI_WARN
from kash.errors import ApiResultError
from kash.kits.media.libs.yt_dlp_tools import ydl_download_media, ydl_extract_info
from kash.model.media_model import SERVICE_APPLE_PODCASTS, MediaMetadata, MediaService, MediaUrlType
from kash.utils.common.type_utils import not_none
from kash.utils.common.url import Url
from kash.utils.file_utils.file_formats_model import MediaType

log = get_logger(__name__)


# URL format is podcast id and episode id:
# https://podcasts.apple.com/us/podcast/upper-priest-lake-trail-to-continental-creek-trail/id1303792223?i=1000394194840
# which is equivalent to
# https://podcasts.apple.com/podcast/id1303792223?i=1000394194840
# See:
# https://podcasters.apple.com/support/847-hosts-and-guests

# mypy is falsely complaining.
# mypy: disable-error-code="override,misc"


class ApplePodcasts(MediaService):
    @override
    def canonicalize_and_type(self, url: Url) -> tuple[Url | None, MediaUrlType | None]:
        parsed_url = urlparse(url)
        if parsed_url.hostname in ("podcasts.apple.com", "itunes.apple.com"):
            path_parts = parsed_url.path.split("/")
            for part in path_parts:
                if part.startswith("id"):
                    podcast_id = part
                    query = parse_qs(parsed_url.query)
                    episode_id = query.get("i", [None])[0]
                    if episode_id:
                        return (
                            Url(f"https://podcasts.apple.com/podcast/{podcast_id}?i={episode_id}"),
                            MediaUrlType.episode,
                        )
                    return (
                        Url(f"https://podcasts.apple.com/podcast/{podcast_id}"),
                        MediaUrlType.podcast,
                    )
        return None, None

    @override
    def get_media_id(self, url: Url) -> str | None:
        parsed_url = urlparse(url)
        if parsed_url.hostname in ("podcasts.apple.com", "itunes.apple.com"):
            path_parts = parsed_url.path.split("/")
            for part in path_parts:
                if part.startswith("id"):
                    podcast_id = part
                    query = parse_qs(parsed_url.query)
                    episode_id = query.get("i", [None])[0]
                    if episode_id:
                        return f"{podcast_id}?i={episode_id}"
        return None

    @override
    def metadata(self, url: Url, full: bool = False) -> MediaMetadata:
        url = not_none(self.canonicalize(url), "Not a recognized Apple Podcasts URL")
        yt_result: dict[str, Any] = self._extract_info(url)

        return self._parse_metadata(yt_result, full=full)

    @override
    def thumbnail_url(self, url: Url) -> Url | None:
        # Apple Podcasts doesn't have a standardized thumbnail URL format.
        # We'll need to extract this from the metadata.
        try:
            metadata = self.metadata(url)
            return metadata.thumbnail_url if metadata else None
        except DownloadError as e:
            log.warning("Could not get a thumbnail URL; will omit thumbnail: %s", e)
            return None

    @override
    def timestamp_url(self, url: Url, timestamp: float) -> Url:
        # Apple Podcasts doesn't support timestamp links. We'll return the original URL.
        return url

    @override
    def download_media(
        self, url: Url, target_dir: Path, media_types: list[MediaType] | None = None
    ) -> dict[MediaType, Path]:
        url = not_none(self.canonicalize(url), "Not a recognized Apple Podcasts URL")
        return ydl_download_media(url, target_dir, media_types)

    def _extract_info(self, url: Url) -> dict[str, Any]:
        url = not_none(self.canonicalize(url), "Not a recognized Apple Podcasts URL")
        return ydl_extract_info(url)

    @override
    def list_channel_items(self, url: Url) -> list[MediaMetadata]:
        result = self._extract_info(url)

        if "entries" in result:
            entries = result["entries"]
        else:
            log.warning("%s No episodes found in the podcast.", EMOJI_WARN)
            entries = []

        episode_meta_list: list[MediaMetadata] = []

        for entry in entries:
            episode_meta_list.append(self._parse_metadata(entry))

        log.message("Found %d episodes in podcast %s", len(episode_meta_list), url)

        return episode_meta_list

    def _parse_metadata(
        self,
        yt_result: dict[str, Any],
        full: bool = False,
        **overrides: dict[str, Any],
    ) -> MediaMetadata:
        try:
            media_id = yt_result["id"]
            if not media_id:
                raise KeyError("No ID found")

            url = yt_result.get("webpage_url") or yt_result.get("url")
            if not url:
                raise KeyError("No URL found")

            thumbnail_url = yt_result.get("thumbnail")

            upload_date_str = yt_result.get("upload_date")
            upload_date = date.fromisoformat(upload_date_str) if upload_date_str else None

            result = MediaMetadata(
                media_id=media_id,
                media_service=SERVICE_APPLE_PODCASTS,
                url=url,
                thumbnail_url=thumbnail_url,
                title=yt_result["title"],
                description=yt_result.get("description"),
                upload_date=upload_date,
                channel_url=Url(yt_result.get("channel_url", "")),
                view_count=yt_result.get("view_count"),
                duration=yt_result.get("duration"),
                heatmap=None,
                **overrides,
            )
            log.message("Parsed Apple Podcasts metadata: %s", result)
        except KeyError as e:
            log.error("Missing key in Apple Podcasts metadata (see saved object): %s", e)
            log.save_object(
                "yt_dlp result", None, to_yaml_string(yt_result, stringify_unknown=True)
            )
            raise ApiResultError("Did not find key in Apple Podcasts metadata: %s" % e)

        return result


## Tests


def test_canonicalize_apple():
    apple = ApplePodcasts()

    assert apple.get_media_id(Url("https://podcasts.apple.com/us/podcast/id1627920305")) is None
    assert apple.get_media_id(Url("https://podcasts.apple.com/podcast/id1234567890")) is None
    assert apple.get_media_id(Url("https://example.com/podcast/123")) is None

    assert (
        apple.get_media_id(Url("https://podcasts.apple.com/podcast/id1234567890?i=1000635337486"))
        == "id1234567890?i=1000635337486"
    )

    def assert_canon(url: str, canon_url: str):
        assert apple.canonicalize(Url(url)) == Url(canon_url)

    assert_canon(
        "https://podcasts.apple.com/us/podcast/redefining-success-money-and-belonging-paul-millerd/id1627920305?i=1000635337486",
        "https://podcasts.apple.com/podcast/id1627920305?i=1000635337486",
    )

    assert_canon(
        "https://podcasts.apple.com/us/podcast/redefining-success-money-and-belonging-paul-millerd/id1627920305?i=1000635337486",
        "https://podcasts.apple.com/podcast/id1627920305?i=1000635337486",
    )



from chopdiff.divs import div
from chopdiff.docs import Paragraph, TextDoc, TextUnit

from kash.config.logger import get_logger
from kash.errors import InvalidInput
from kash.exec import kash_action, kash_precondition
from kash.exec.llm_transforms import llm_transform_str
from kash.model import Format, Item, ItemType, LLMOptions, Message, MessageTemplate
from kash.utils.common.task_stack import task_stack

log = get_logger(__name__)


llm_options = LLMOptions(
    system_message=Message(
        """
        You are a careful and precise editor.
        You give exactly the results requested without additional commentary.
        """
    ),
    body_template=MessageTemplate(
        """
        Please describe what is said in the following one or two paragraphs, as a 
        summary or caption for the content. Rules:

        - Mention only the most important points. Include all the key topics discussed.
        
        - Keep the caption short! Use one or two short, complete sentences. It must be significantly
            shorter than the input text.
        
        - Write in clean and and direct language.

        - Do not mention the text or the author. Simply state the points as presented.

        - If the content contains other promotional material or only references information such as
            about what will be discussed later, ignore it.

        - DO NOT INCLUDE any other commentary.

        - If the input is very short or so unclear you can't summarize it, simply output
            "(No results)".

        Sample input text:

        I think push ups are one of the most underrated exercises out there and they're also one of
        the exercises that is most frequently performed with poor technique.
        And I think this is because a lot of people think it's just an easy exercise and they adopt
        a form that allows them to achieve a rep count that they would expect from an easy exercise,
        but all that ends up happening is they they do a bunch of poor quality repetitions in order
        to get a high rep count. So I don't think push ups are particularly easy when they're done well
        and they're really effective for building just general fitness and muscle in the upper body
        if you do them properly. So here's how you get the most out of them.

        Sample output text:

        Push ups are an underrated exercise. They are not easy to do well so are often done poorly.

        Input text:

        {body}

        Output text:
        """
    ),
)


PARA = "para"
ANNOTATED_PARA = "annotated-para"
PARA_CAPTION = "para-caption"


@kash_precondition
def has_annotated_paras(item: Item) -> bool:
    """
    Useful to check if an item has already been annotated with captions.
    """
    return bool(item.body and item.body.find(f'<p class="{ANNOTATED_PARA}">') != -1)


def process_para(llm_options: LLMOptions, para: Paragraph) -> str:
    caption_div = None

    para_str = para.reassemble()
    # Only caption actual paragraphs with enough words.
    if not para.is_markup() and not para.is_header() and para.size(TextUnit.words) > 40:
        log.message("Captioning paragraph (%s words)", para.size(TextUnit.words))
        llm_response = llm_transform_str(llm_options, para_str)
        caption_div = div(PARA_CAPTION, llm_response)
        new_div = div(ANNOTATED_PARA, caption_div, div(PARA, para_str))
    else:
        log.message(
            "Skipping captioning very short paragraph (%s words)", para.size(TextUnit.words)
        )
        new_div = para_str
    return new_div


@kash_action(llm_options=llm_options)
def caption_paras(item: Item) -> Item:
    """
    Caption each paragraph in the text with a very short summary, wrapping the original
    and the caption in simple divs.
    """
    if not item.body:
        raise InvalidInput(f"Item must have a body: {item}")

    doc = TextDoc.from_text(item.body)
    output = []
    with task_stack().context("caption_paras", doc.size(TextUnit.paragraphs), "para") as ts:
        for para in doc.paragraphs:
            if para.size(TextUnit.words) > 0:
                output.append(process_para(llm_options, para))
            ts.next()

    final_output = "\n\n".join(output)
    result_item = item.derived_copy(type=ItemType.doc, body=final_output, format=Format.md_html)

    return result_item



from kash.exec import kash_action, llm_transform_item
from kash.model import Item, LLMOptions, Message, MessageTemplate, TitleTemplate

llm_options = LLMOptions(
    system_message=Message(
        """
        You are a careful and precise editor.
        You give exactly the results requested without additional commentary.
        """
    ),
    # TODO: Get a scene context/description from the resource and add it to this template.
    # Is this a book/article/transcript/lecture/interview/etc? Who was involved? When was it?
    # Create a question template and have this filled in.
    body_template=MessageTemplate(
        """
        Give a brief description of the entire text below, as a summary of two or three sentences.
        Write it concisely and clearly, in a form suitable for a short description of a web page
        or article.

        - Use simple and precise language.

        - Simply state the facts or claims without referencing the text or the author. For example, if the
          text is about cheese being nutritious, you can say "Cheese is nutritious." But do NOT
          say "The author says cheese is nutritious" or "According to the text, cheese is nutritious."

        - If the content is missing so brief that it can't be described, simply say "(No description.)"
        
        Original text:

        {body}

        Brief description of the text:
        """
    ),
)


@kash_action(llm_options=llm_options, title_template=TitleTemplate("Summary of {title}"))
def describe_briefly(item: Item) -> Item:
    """
    Write a brief description of a text, in at most three sentences.
    """
    return llm_transform_item(item)
